/**
 * Copyright 2005 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.nutch.indexer.extra;

import org.apache.commons.jexl.Expression;
import org.apache.commons.jexl.ExpressionFactory;
import org.apache.commons.jexl.JexlContext;
import org.apache.commons.jexl.JexlHelper;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;

import org.apache.nutch.parse.Parse;

import org.apache.nutch.indexer.IndexingFilter;
import org.apache.nutch.indexer.IndexingException;

import org.apache.nutch.crawl.CrawlDatum;
import org.apache.nutch.crawl.Inlinks;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.UTF8;

import java.util.List;

import org.apache.regexp.RE;

/**
 * The index-extra filter allows the configuration of additional document fields, which can be created
 * from the document text, from meta data or from previously created document fields.
 * 
 * Known Issues
 * ------------
 * 1)  For this plugin to work correctly on any document field, it is necessary to run the other index filters
 * first, so that all basic document fields are generated first.  To do this, configure the indexingfilter.order
 * property.
 * (PLEASE SEE PATCH NUTCH-421 TO ENABLE indexingfilter.order property)
 * If this patch is not applied, the plugin will still work, but will not be able to use document fields created
 * by other index filter plugins.
 * 2)  At this stage, field boost can not be used as Nutch scoring overrides the field boost with its own
 * document-level boost calculation.  This occurs at the end of org.apache.nutch.indexer.Indexer's reduce method.
 *
 * @author Alan Tanaman (alan.tanaman NOSPAM-AT-NOSPAM idna-solutions.com)
 */

public class ExtraIndexingFilter implements IndexingFilter {
  public static final Log LOG = LogFactory.getLog(ExtraIndexingFilter.class);

  public Document filter(Document doc, Parse parse, UTF8 url, CrawlDatum datum, Inlinks inlinks)
    throws IndexingException {

	  /* Get configuration file properties into a list */
	  List extraIndexingMetaDataList = ExtraIndexingFilterConfig.getInstance().getExtraIndexerProperties();
	  
	  /* Potentially add a document field for each //extraIndexingMetaData node in the list */
	  if (extraIndexingMetaDataList != null && extraIndexingMetaDataList.size() > 0)
	  {
			for (int j = 0; j < extraIndexingMetaDataList.size(); j++)
			{
				ExtraIndexingMetaData extraIndexingMetaData = (ExtraIndexingMetaData)extraIndexingMetaDataList.get(j);
				/* A field is created if the node is unconditional or if the condition is true */
				if (isTrue(doc, parse, extraIndexingMetaData.getCondition()))
				{
					addField(doc, parse, extraIndexingMetaData);
				}
			}		  
	  }

    return doc;
  }

  private Boolean isTrue (Document doc, Parse parse, String condition)
  {
	  /************************************************************************
	   * Check whether the condition stated in the condition attribute is true
	   ************************************************************************/

	  /* If no or zero-length condition attribute, then assume true */
	  if(condition == null || condition.length() == 0)
          return true;

	  /* Create a Jexl context with variables that can be used by the expression */	  
	  JexlContext jc = JexlHelper.createContext();
      jc.getVars().put("document", doc);
      jc.getVars().put("meta", parse.getData());
      jc.getVars().put("text", parse.getText());
      
      try
      {
    	  /* Create a Jexl expression from the condition and return its value */	  
          Expression exp = ExpressionFactory.createExpression(condition);
          return ((Boolean)exp.evaluate(jc)).booleanValue();
      }
      catch (Exception e)
      {
          e.printStackTrace();
          return false;
      }
  }

  private Document addField(Document doc, Parse parse, ExtraIndexingMetaData extraIndexingMetaData )
  {
	  /************************************************************************
	   * Add a field to the document
	   ************************************************************************/
	  
	  String fieldValue = applyRegex(getSourceValue(doc, parse, extraIndexingMetaData), extraIndexingMetaData);
			  
	  if (fieldValue != null)
	  {
		  Field.Index indexType;
		  String indexed = extraIndexingMetaData.getIndexed();
		  if (indexed.equals("TOKENIZED")) {
			  indexType = Field.Index.TOKENIZED;
		  } else if (indexed.equals("UN_TOKENIZED")) {
			  indexType = Field.Index.UN_TOKENIZED;
		  } else if (indexed.equals("NO")) {
			  indexType = Field.Index.NO;
		  } else {
			  indexType = Field.Index.TOKENIZED;
        	  LOG.error ("Invalid indexed attribute.  Valid values are 'tokenized', 'un_tokenized' or 'no'");
		  }
		  
		  Field.Store storeType;
		  String stored = extraIndexingMetaData.getStored();
		  if (stored.equals("YES")) {
			  storeType = Field.Store.YES;
		  } else if (stored.equals("COMPRESS")) {
			  storeType = Field.Store.COMPRESS;
		  } else if (stored.equals("NO")) {
			  storeType = Field.Store.NO;
		  }	else {
			  storeType = Field.Store.YES;
        	  LOG.error ("Invalid stored attribute.  Valid values are 'yes', 'compress' or 'no'");
		  }
		  
		  Field.TermVector termVectorType;
		  String termVector = extraIndexingMetaData.getTermVector();
		  if (termVector.equals("YES")) {
			  termVectorType = Field.TermVector.YES;
	      } else if (termVector.equals("WITH_OFFSETS")) {
	    	  termVectorType = Field.TermVector.WITH_OFFSETS;
	      } else if (termVector.equals("WITH_POSITIONS")) {
	    	  termVectorType = Field.TermVector.WITH_POSITIONS;
	      } else if (termVector.equals("WITH_POSITIONS_OFFSETS")) {
	    	  termVectorType = Field.TermVector.WITH_POSITIONS_OFFSETS;
		  } else if (termVector.equals("NO")) {
			  termVectorType = Field.TermVector.NO;
		  } else {
			  termVectorType = Field.TermVector.NO;
        	  LOG.error ("Invalid termVector attribute.  Valid values are 'yes', 'with_positions', 'with_offsets', 'with_positions_offsets' or 'no'");
		  }
		  
		  Field field = new Field(extraIndexingMetaData.getName(),
	    		  fieldValue,
	    		  storeType,
	    		  indexType,
	    		  termVectorType);
		  if (!extraIndexingMetaData.getBoost().equals("1.0")) {
			  field.setBoost(Float.parseFloat(extraIndexingMetaData.getBoost()));
		  }
		  LOG.info ("Adding field " + extraIndexingMetaData.getName());
		  doc.add(field);
	  }
    return doc;
  }

  private String getSourceValue(Document doc, Parse parse, ExtraIndexingMetaData extraIndexingMetaData )
  {
	  /************************************************************************
	   * Extract the  sourceValue from the relevant source depending upon sourceType
	   ************************************************************************/

	  switch(extraIndexingMetaData.getSourceType()) {
      case CONST: {
    	  /* The source value is a constant string in the sourceValue attribute */	  
    	  return extraIndexingMetaData.getSourceValue();
      }
      case FIELD: {
    	  /* The source value will be the document field named in the sourceValue attribute */	  
    	  return doc.get(extraIndexingMetaData.getSourceValue());
      }
      case JAVA: {
    	  /* The source value will be the result of the expression given in the sourceValue attribute */	  

    	  /* Create a Jexl context with variables that can be used by the expression */	  
    	  JexlContext jc = JexlHelper.createContext();
          jc.getVars().put("document", doc);
          jc.getVars().put("data", parse.getData());
          jc.getVars().put("text", parse.getText());

          try {
        	  /* Create a Jexl expression from the sourceValue attribute and return its value */	  
        	  Expression exp = ExpressionFactory.createExpression(extraIndexingMetaData.getSourceValue());
	          try {
	        	  return exp.evaluate(jc).toString();
	          }
	          catch (Exception e) {
	        	  LOG.error ("Exception evaluating expression: " + extraIndexingMetaData.getSourceValue());
	        	  LOG.error ("Exception stack trace: " + e);
	        	  return null;
	          }
          }
          catch (Exception e) {
        	  LOG.error ("Exception incorrect JEXL expression: " + extraIndexingMetaData.getSourceValue());
        	  LOG.error ("Exception stack trace: " + e);
        	  return null;
          }
      }
      case META: {
    	  /* The source value will be the meta data field named in the sourceValue attribute */	  
    	  return parse.getData().getMeta(extraIndexingMetaData.getSourceValue());
      }
      case TEXT: {
    	  /* The source value will be the parsed text */	  
    	  return parse.getText();
      }
      default: {
	  // TODO: Error handling
    	  System.out.println(" invalid source type"); 
    	  return null;
	  }
      }
  }
  
  private String applyRegex (String sourceValue, ExtraIndexingMetaData extraIndexingMetaData)
  {
	  /************************************************************************
	   * If a regex is to be applied to the sourceValue, check whether it matches
	   ************************************************************************/

	  if (sourceValue != null)
	  {
		  String regexValue = extraIndexingMetaData.getRegexValue();
		  if (regexValue != null)
		  {
        	  /* Create a regex based on the regexValue attribute */
			  try {
		          RE re = new RE(regexValue);
	        	  /* Apply the regex to the sourceValue */	  
		          if(re.match(sourceValue))
		        	  /* Match: Return the value the regex placed in parenthesis #regexNo */	  
		        	  return re.getParen(extraIndexingMetaData.getRegexNo());
		          else
		        	  /* No match: Return null */	  
		        	  return null;
			  } catch (Exception e) {
	        	  LOG.error ("Exception incorrect regex: " + regexValue);
	        	  LOG.error ("Exception stack trace: " + e);
	        	  return null;
			  }
		  }
	  }
	  /* No need to apply regex - return same value received */	  
	  return sourceValue;
  }

  private Configuration conf;

  public void setConf(Configuration conf) {
    this.conf = conf;
  }

  public Configuration getConf() {
    return this.conf;
  }

}